@@ -1,4 +1,4 @@
-/*	$OpenBSD: control.c,v 1.15 2006/04/25 08:00:22 claudio Exp $ */
+/*	$OpenBSD: control.c,v 1.18 2007/03/21 10:54:30 claudio Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -197,7 +197,7 @@ control_dispatch_imsg(int fd, short event, void *bula)
 
 	switch (event) {
 	case EV_READ:
-		if ((n = imsg_read(&c->ibuf)) <= 0) {
+		if ((n = imsg_read(&c->ibuf)) == -1 || n == 0) {
 			control_close(fd);
 			return;
 		}
@@ -225,6 +225,8 @@ control_dispatch_imsg(int fd, short event, void *bula)
 		switch (imsg.hdr.type) {
 		case IMSG_CTL_FIB_COUPLE:
 		case IMSG_CTL_FIB_DECOUPLE:
+			ospfe_fip_update(imsg.hdr.type);
+			/* FALLTHROUGH */
 		case IMSG_CTL_RELOAD:
 			c->ibuf.pid = imsg.hdr.pid;
 			ospfe_imsg_compose_parent(imsg.hdr.type, 0, NULL, 0);
@@ -291,7 +293,7 @@ session_socket_blockmode(int fd, enum blockmodes bm)
 	int	flags;
 
 	if ((flags = fcntl(fd, F_GETFL, 0)) == -1)
-		fatal("fnctl F_GETFL");
+		fatal("fcntl F_GETFL");
 
 	if (bm == BM_NONBLOCK)
 		flags |= O_NONBLOCK;
@@ -299,5 +301,5 @@ session_socket_blockmode(int fd, enum blockmodes bm)
 		flags &= ~O_NONBLOCK;
 
 	if ((flags = fcntl(fd, F_SETFL, flags)) == -1)
-		fatal("fnctl F_SETFL");
+		fatal("fcntl F_SETFL");
 }
