@@ -1,4 +1,4 @@
-/*	$OpenBSD: interface.c,v 1.52 2006/08/18 11:54:28 claudio Exp $ */
+/*	$OpenBSD: interface.c,v 1.59 2007/05/29 22:08:25 claudio Exp $ */
 
 /*
  * Copyright (c) 2005 Claudio Jeker <claudio@openbsd.org>
@@ -139,6 +139,13 @@ if_fsm(struct iface *iface, enum iface_event event)
 	if (iface->state != old_state)
 		orig_rtr_lsa(iface->area);
 
+	if (old_state & (IF_STA_MULTI | IF_STA_POINTTOPOINT) &&
+	    (iface->state & (IF_STA_MULTI | IF_STA_POINTTOPOINT)) == 0)
+		ospfe_demote_iface(iface, 0);
+	if ((old_state & (IF_STA_MULTI | IF_STA_POINTTOPOINT)) == 0 &&
+	    iface->state & (IF_STA_MULTI | IF_STA_POINTTOPOINT))
+		ospfe_demote_iface(iface, 1);
+
 	log_debug("if_fsm: event %s resulted in action %s and changing "
 	    "state for interface %s from %s to %s",
 	    if_event_names[event], if_action_names[iface_fsm[i].action],
@@ -148,12 +155,9 @@ if_fsm(struct iface *iface, enum iface_event event)
 }
 
 struct iface *
-if_new(struct kif *kif)
+if_new(struct kif *kif, struct kif_addr *ka)
 {
-	struct sockaddr_in	*sain;
 	struct iface		*iface;
-	struct ifreq		*ifr;
-	int			 s;
 
 	if ((iface = calloc(1, sizeof(*iface))) == NULL)
 		err(1, "if_new: calloc");
@@ -177,14 +181,6 @@ if_new(struct kif *kif)
 
 	strlcpy(iface->name, kif->ifname, sizeof(iface->name));
 
-	if ((ifr = calloc(1, sizeof(*ifr))) == NULL)
-		err(1, "if_new: calloc");
-
-	/* set up ifreq */
-	strlcpy(ifr->ifr_name, kif->ifname, sizeof(ifr->ifr_name));
-	if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
-		err(1, "if_new: socket");
-
 	/* get type */
 	if (kif->flags & IFF_POINTOPOINT)
 		iface->type = IF_TYPE_POINTOPOINT;
@@ -203,29 +199,13 @@ if_new(struct kif *kif)
 	iface->linkstate = kif->link_state;
 	iface->media_type = kif->media_type;
 
-	/* get address */
-	if (ioctl(s, SIOCGIFADDR, ifr) < 0)
-		err(1, "if_new: cannot get address");
-	sain = (struct sockaddr_in *)&ifr->ifr_addr;
-	iface->addr = sain->sin_addr;
-
-	/* get mask */
-	if (ioctl(s, SIOCGIFNETMASK, ifr) < 0)
-		err(1, "if_new: cannot get mask");
-	sain = (struct sockaddr_in *)&ifr->ifr_addr;
-	iface->mask = sain->sin_addr;
-
-	/* get p2p dst address */
+	/* set address, mask and p2p addr */
+	iface->addr = ka->addr;
+	iface->mask = ka->mask;
 	if (kif->flags & IFF_POINTOPOINT) {
-		if (ioctl(s, SIOCGIFDSTADDR, ifr) < 0)
-			err(1, "if_new: cannot get dst addr");
-		sain = (struct sockaddr_in *)&ifr->ifr_addr;
-		iface->dst = sain->sin_addr;
+		iface->dst = ka->dstbrd;
 	}
 
-	free(ifr);
-	close(s);
-
 	return (iface);
 }
 
@@ -236,10 +216,21 @@ if_del(struct iface *iface)
 
 	log_debug("if_del: interface %s", iface->name);
 
+	/* revert the demotion when the interface is deleted */
+	if ((iface->state & (IF_STA_MULTI | IF_STA_POINTTOPOINT)) == 0)
+		ospfe_demote_iface(iface, 1);
+
 	/* clear lists etc */
 	while ((nbr = LIST_FIRST(&iface->nbr_list)) != NULL)
 		nbr_del(nbr);
 
+	if (evtimer_pending(&iface->hello_timer, NULL))
+		evtimer_del(&iface->hello_timer);
+	if (evtimer_pending(&iface->wait_timer, NULL))
+		evtimer_del(&iface->wait_timer);
+	if (evtimer_pending(&iface->lsack_tx_timer, NULL))
+		evtimer_del(&iface->lsack_tx_timer);
+
 	ls_ack_list_clr(iface);
 	md_list_clr(&iface->auth_md_list);
 	free(iface);
@@ -257,6 +248,8 @@ if_init(struct ospfd_conf *xconf, struct iface *iface)
 	evtimer_set(&iface->wait_timer, if_wait_timer, iface);
 
 	iface->fd = xconf->ospf_socket;
+
+	ospfe_demote_iface(iface, 0);
 }
 
 /* timers */
@@ -328,7 +321,7 @@ if_act_start(struct iface *iface)
 	struct timeval		 now;
 
 	if (!((iface->flags & IFF_UP) &&
-	    (iface->linkstate == LINK_STATE_UP ||
+	    (LINK_STATE_IS_UP(iface->linkstate) ||
 	    (iface->linkstate == LINK_STATE_UNKNOWN &&
 	    iface->media_type != IFT_CARP)))) {
 		log_debug("if_act_start: interface %s link down",
@@ -355,11 +348,8 @@ if_act_start(struct iface *iface)
 	switch (iface->type) {
 	case IF_TYPE_POINTOPOINT:
 		inet_aton(AllSPFRouters, &addr);
-		if (if_join_group(iface, &addr)) {
-			log_warnx("if_act_start: error joining group %s, "
-			    "interface %s", inet_ntoa(addr), iface->name);
+		if (if_join_group(iface, &addr))
 			return (-1);
-		}
 		iface->state = IF_STA_POINTTOPOINT;
 		break;
 	case IF_TYPE_VIRTUALLINK:
@@ -372,11 +362,8 @@ if_act_start(struct iface *iface)
 		return (-1);
 	case IF_TYPE_BROADCAST:
 		inet_aton(AllSPFRouters, &addr);
-		if (if_join_group(iface, &addr)) {
-			log_warnx("if_act_start: error joining group %s, "
-			    "interface %s", inet_ntoa(addr), iface->name);
+		if (if_join_group(iface, &addr))
 			return (-1);
-		}
 		if (iface->priority == 0) {
 			iface->state = IF_STA_DROTHER;
 		} else {
@@ -513,20 +500,12 @@ start:
 		inet_aton(AllDRouters, &addr);
 		if (old_state & IF_STA_DRORBDR &&
 		    (iface->state & IF_STA_DRORBDR) == 0) {
-			if (if_leave_group(iface, &addr)) {
-				log_warnx("if_act_elect: "
-				    "error leaving group %s, interface %s",
-				    inet_ntoa(addr), iface->name);
+			if (if_leave_group(iface, &addr))
 				return (-1);
-			}
 		} else if ((old_state & IF_STA_DRORBDR) == 0 &&
 		    iface->state & IF_STA_DRORBDR) {
-			if (if_join_group(iface, &addr)) {
-				log_warnx("if_act_elect: "
-				    "error joining group %s, interface %s",
-				    inet_ntoa(addr), iface->name);
+			if (if_join_group(iface, &addr))
 				return (-1);
-			}
 		}
 
 		LIST_FOREACH(nbr, &iface->nbr_list, entry) {
@@ -599,6 +578,9 @@ if_act_reset(struct iface *iface)
 	if_stop_hello_timer(iface);
 	if_stop_wait_timer(iface);
 
+	/* send empty hello to tell everybody that we are going down */
+	send_hello(iface);
+
 	return (0);
 }
 
@@ -674,27 +656,13 @@ if_to_ctl(struct iface *iface)
 
 /* misc */
 int
-if_set_mcast_ttl(int fd, u_int8_t ttl)
+if_set_recvif(int fd, int enable)
 {
-	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL,
-	    (char *)&ttl, sizeof(ttl)) < 0) {
-		log_warn("if_set_mcast_ttl: error setting "
-		    "IP_MULTICAST_TTL to %d", ttl);
+	if (setsockopt(fd, IPPROTO_IP, IP_RECVIF, &enable,
+	    sizeof(enable)) < 0) {
+		log_warn("if_set_recvif: error setting IP_RECVIF");
 		return (-1);
 	}
-
-	return (0);
-}
-
-int
-if_set_tos(int fd, int tos)
-{
-	if (setsockopt(fd, IPPROTO_IP, IP_TOS,
-	    (int *)&tos, sizeof(tos)) < 0) {
-		log_warn("if_set_tos: error setting IP_TOS to 0x%x", tos);
-		return (-1);
-	}
-
 	return (0);
 }
 
@@ -709,21 +677,52 @@ if_set_recvbuf(int fd)
 		bsize /= 2;
 }
 
+/*
+ * only one JOIN or DROP per interface and address is allowed so we need
+ * to keep track of what is added and removed.
+ */
+struct if_group_count {
+	LIST_ENTRY(if_group_count)	entry;
+	struct in_addr			addr;
+	unsigned int			ifindex;
+	int				count;
+};
+
+LIST_HEAD(,if_group_count) ifglist = LIST_HEAD_INITIALIZER(ifglist);
+
 int
 if_join_group(struct iface *iface, struct in_addr *addr)
 {
-	struct ip_mreq	 mreq;
+	struct ip_mreq		 mreq;
+	struct if_group_count	*ifg;
 
 	switch (iface->type) {
 	case IF_TYPE_POINTOPOINT:
 	case IF_TYPE_BROADCAST:
+		LIST_FOREACH(ifg, &ifglist, entry)
+			if (iface->ifindex == ifg->ifindex &&
+			    addr->s_addr == ifg->addr.s_addr)
+				break;
+		if (ifg == NULL) {
+			if ((ifg = calloc(1, sizeof(*ifg))) == NULL)
+				fatal("if_join_group");
+			ifg->addr.s_addr = addr->s_addr;
+			ifg->ifindex = iface->ifindex;
+			LIST_INSERT_HEAD(&ifglist, ifg, entry);
+		}
+
+		if (ifg->count++ != 0)
+			/* already joined */
+			return (0);
+
 		mreq.imr_multiaddr.s_addr = addr->s_addr;
 		mreq.imr_interface.s_addr = iface->addr.s_addr;
 
 		if (setsockopt(iface->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
 		    (void *)&mreq, sizeof(mreq)) < 0) {
-			log_debug("if_join_group: error IP_ADD_MEMBERSHIP, "
-			    "interface %s", iface->name);
+			log_warn("if_join_group: error IP_ADD_MEMBERSHIP, "
+			    "interface %s address %s", iface->name,
+			    inet_ntoa(*addr));
 			return (-1);
 		}
 		break;
@@ -743,20 +742,37 @@ if_join_group(struct iface *iface, struct in_addr *addr)
 int
 if_leave_group(struct iface *iface, struct in_addr *addr)
 {
-	struct ip_mreq	 mreq;
+	struct ip_mreq		 mreq;
+	struct if_group_count	*ifg;
 
 	switch (iface->type) {
 	case IF_TYPE_POINTOPOINT:
 	case IF_TYPE_BROADCAST:
+		LIST_FOREACH(ifg, &ifglist, entry)
+			if (iface->ifindex == ifg->ifindex &&
+			    addr->s_addr == ifg->addr.s_addr)
+				break;
+
+		/* if interface is not found just try to drop membership */
+		if (ifg && --ifg->count != 0)
+			/* others still joined */
+			return (0);
+
 		mreq.imr_multiaddr.s_addr = addr->s_addr;
 		mreq.imr_interface.s_addr = iface->addr.s_addr;
 
 		if (setsockopt(iface->fd, IPPROTO_IP, IP_DROP_MEMBERSHIP,
 		    (void *)&mreq, sizeof(mreq)) < 0) {
-			log_debug("if_leave_group: error IP_DROP_MEMBERSHIP, "
-			    "interface %s", iface->name);
+			log_warn("if_leave_group: error IP_DROP_MEMBERSHIP, "
+			    "interface %s address %s", iface->name,
+			    inet_ntoa(*addr));
 			return (-1);
 		}
+
+		if (ifg) {
+			LIST_REMOVE(ifg, entry);
+			free(ifg);
+		}
 		break;
 	case IF_TYPE_POINTOMULTIPOINT:
 	case IF_TYPE_VIRTUALLINK:
@@ -810,3 +826,16 @@ if_set_mcast_loop(int fd)
 
 	return (0);
 }
+
+int
+if_set_ip_hdrincl(int fd)
+{
+	int	hincl = 1;
+
+	if (setsockopt(fd, IPPROTO_IP, IP_HDRINCL, &hincl, sizeof(hincl)) < 0) {
+		log_warn("if_set_ip_hdrincl: error setting IP_HDRINCL");
+		return (-1);
+	}
+
+	return (0);
+}
