@@ -1,4 +1,4 @@
-/*	$OpenBSD: neighbor.c,v 1.33 2006/08/06 12:36:23 claudio Exp $ */
+/*	$OpenBSD: neighbor.c,v 1.35 2007/06/13 17:47:52 claudio Exp $ */
 
 /*
  * Copyright (c) 2005 Claudio Jeker <claudio@openbsd.org>
@@ -111,7 +111,8 @@ const char * const nbr_action_names[] = {
 	"ADJ_OK",
 	"RESET_DD",
 	"DELETE",
-	"CLEAR_LISTS"
+	"CLEAR_LISTS",
+	"HELLO_CHK"
 };
 
 int
@@ -325,6 +326,17 @@ nbr_del(struct nbr *nbr)
 {
 	ospfe_imsg_compose_rde(IMSG_NEIGHBOR_DOWN, nbr->peerid, 0, NULL, 0);
 
+	if (evtimer_pending(&nbr->inactivity_timer, NULL))
+		evtimer_del(&nbr->inactivity_timer);
+	if (evtimer_pending(&nbr->db_tx_timer, NULL))
+		evtimer_del(&nbr->db_tx_timer);
+	if (evtimer_pending(&nbr->lsreq_tx_timer, NULL))
+		evtimer_del(&nbr->lsreq_tx_timer);
+	if (evtimer_pending(&nbr->ls_retrans_timer, NULL))
+		evtimer_del(&nbr->ls_retrans_timer);
+	if (evtimer_pending(&nbr->adj_timer, NULL))
+		evtimer_del(&nbr->adj_timer);
+
 	/* clear lists */
 	ls_retrans_list_clr(nbr);
 	db_sum_list_clr(nbr);
