@@ -1,4 +1,4 @@
-/*	$OpenBSD: ospfctl.c,v 1.35 2006/08/23 08:26:03 claudio Exp $ */
+/*	$OpenBSD: ospfctl.c,v 1.37 2007/05/30 02:21:08 henning Exp $ */
 
 /*
  * Copyright (c) 2005 Claudio Jeker <claudio@openbsd.org>
@@ -338,8 +338,6 @@ get_ifms_type(int mediatype)
 		return (IFM_ETHER);
 	case IFT_FDDI:
 		return (IFM_FDDI);
-	case IFT_ISO88025:
-		return (IFM_TOKEN);
 	case IFT_CARP:
 		return (IFM_CARP);
 	case IFT_PPP:
@@ -1198,7 +1196,9 @@ get_linkstate(int media_type, int link_state)
 			if (p->ifms_type != media_type ||
 			    p->ifms_valid != ifm_status_valid_list[i])
 				continue;
-			return (p->ifms_string[link_state == LINK_STATE_UP]);
+			if (LINK_STATE_IS_UP(link_state))
+				return (p->ifms_string[1]);
+			return (p->ifms_string[0]);
 		}
 
 	return ("unknown");
@@ -1235,9 +1235,6 @@ show_fib_interface_msg(struct imsg *imsg)
 		case IFT_FDDI:
 			ifms_type = IFM_FDDI;
 			break;
-		case IFT_ISO88025:
-			ifms_type = IFM_TOKEN;
-			break;
 		case IFT_CARP:
 			ifms_type = IFM_CARP;
 			break;
