@@ -1,4 +1,4 @@
-/*	$OpenBSD: ospfe.c,v 1.47 2006/06/02 18:49:55 norby Exp $ */
+/*	$OpenBSD: ospfe.c,v 1.57 2007/07/25 19:11:27 claudio Exp $ */
 
 /*
  * Copyright (c) 2005 Claudio Jeker <claudio@openbsd.org>
@@ -48,9 +48,10 @@ void		 ospfe_shutdown(void);
 void		 orig_rtr_lsa_all(struct area *);
 struct iface	*find_vlink(struct abr_rtr *);
 
-struct ospfd_conf	*oeconf = NULL;
+struct ospfd_conf	*oeconf = NULL, *nconf;
 struct imsgbuf		*ibuf_main;
 struct imsgbuf		*ibuf_rde;
+int			 oe_nofib;
 
 /* ARGSUSED */
 void
@@ -71,8 +72,8 @@ pid_t
 ospfe(struct ospfd_conf *xconf, int pipe_parent2ospfe[2], int pipe_ospfe2rde[2],
     int pipe_parent2rde[2])
 {
-	struct area	*area = NULL;
-	struct iface	*iface = NULL;
+	struct area	*area;
+	struct iface	*iface;
 	struct redistribute *r;
 	struct passwd	*pw;
 	struct event	 ev_sigint, ev_sigterm;
@@ -97,19 +98,17 @@ ospfe(struct ospfd_conf *xconf, int pipe_parent2ospfe[2], int pipe_ospfe2rde[2],
 		fatal("error creating raw socket");
 
 	/* set some defaults */
-	if (if_set_mcast_ttl(xconf->ospf_socket,
-	    IP_DEFAULT_MULTICAST_TTL) == -1)
-		fatal("if_set_mcast_ttl");
-
 	if (if_set_mcast_loop(xconf->ospf_socket) == -1)
 		fatal("if_set_mcast_loop");
-
-	if (if_set_tos(xconf->ospf_socket, IPTOS_PREC_INTERNETCONTROL) == -1)
-		fatal("if_set_tos");
-
+	if (if_set_ip_hdrincl(xconf->ospf_socket) == -1)
+		fatal("if_set_ip_hdrincl");
+	if (if_set_recvif(xconf->ospf_socket, 1) == -1)
+		fatal("if_set_recvif");
 	if_set_recvbuf(xconf->ospf_socket);
 
 	oeconf = xconf;
+	if (oeconf->flags & OSPFD_FLAG_NO_FIB_UPDATE)
+		oe_nofib = 1;
 
 	if ((pw = getpwnam(OSPFD_USER)) == NULL)
 		fatal("getpwnam");
@@ -137,6 +136,7 @@ ospfe(struct ospfd_conf *xconf, int pipe_parent2ospfe[2], int pipe_ospfe2rde[2],
 	signal_add(&ev_sigint, NULL);
 	signal_add(&ev_sigterm, NULL);
 	signal(SIGPIPE, SIG_IGN);
+	signal(SIGHUP, SIG_IGN);
 
 	/* setup pipes */
 	close(pipe_parent2ospfe[0]);
@@ -180,6 +180,7 @@ ospfe(struct ospfd_conf *xconf, int pipe_parent2ospfe[2], int pipe_ospfe2rde[2],
 
 	/* start interfaces */
 	LIST_FOREACH(area, &oeconf->area_list, entry) {
+		ospfe_demote_area(area, 0);
 		LIST_FOREACH(iface, &area->iface_list, entry) {
 			if_init(xconf, iface);
 			if (if_fsm(iface, IF_EVT_UP)) {
@@ -248,19 +249,22 @@ ospfe_imsg_compose_rde(int type, u_int32_t peerid, pid_t pid,
 void
 ospfe_dispatch_main(int fd, short event, void *bula)
 {
+	static struct area	*narea;
+	static struct iface	*niface;
 	struct imsg	 imsg;
 	struct imsgbuf  *ibuf = bula;
 	struct area	*area = NULL;
 	struct iface	*iface = NULL;
 	struct kif	*kif;
-	int		 n, link_ok;
+	struct auth_md	 md;
+	int		 n, link_ok, stub_changed, shut = 0;
 
 	switch (event) {
 	case EV_READ:
 		if ((n = imsg_read(ibuf)) == -1)
 			fatal("imsg_read error");
 		if (n == 0)	/* connection closed */
-			fatalx("pipe closed");
+			shut = 1;
 		break;
 	case EV_WRITE:
 		if (msgbuf_write(&ibuf->w) == -1)
@@ -284,7 +288,7 @@ ospfe_dispatch_main(int fd, short event, void *bula)
 				fatalx("IFINFO imsg with wrong len");
 			kif = imsg.data;
 			link_ok = (kif->flags & IFF_UP) &&
-			    (kif->link_state == LINK_STATE_UP ||
+			    (LINK_STATE_IS_UP(kif->link_state) ||
 			    (kif->link_state == LINK_STATE_UNKNOWN &&
 			    kif->media_type != IFT_CARP));
 
@@ -314,6 +318,53 @@ ospfe_dispatch_main(int fd, short event, void *bula)
 				}
 			}
 			break;
+		case IMSG_RECONF_CONF:
+			if ((nconf = malloc(sizeof(struct ospfd_conf))) ==
+			    NULL)
+				fatal(NULL);
+			memcpy(nconf, imsg.data, sizeof(struct ospfd_conf));
+
+			LIST_INIT(&nconf->area_list);
+			LIST_INIT(&nconf->cand_list);
+			break;
+		case IMSG_RECONF_AREA:
+			if ((narea = area_new()) == NULL)
+				fatal(NULL);
+			memcpy(narea, imsg.data, sizeof(struct area));
+
+			LIST_INIT(&narea->iface_list);
+			LIST_INIT(&narea->nbr_list);
+			RB_INIT(&narea->lsa_tree);
+
+			LIST_INSERT_HEAD(&nconf->area_list, narea, entry);
+			break;
+		case IMSG_RECONF_IFACE:
+			if ((niface = malloc(sizeof(struct iface))) == NULL)
+				fatal(NULL);
+			memcpy(niface, imsg.data, sizeof(struct iface));
+
+			LIST_INIT(&niface->nbr_list);
+			TAILQ_INIT(&niface->ls_ack_list);
+			TAILQ_INIT(&niface->auth_md_list);
+
+			niface->area = narea;
+			LIST_INSERT_HEAD(&narea->iface_list, niface, entry);
+			break;
+		case IMSG_RECONF_AUTHMD:
+			memcpy(&md, imsg.data, sizeof(struct auth_md));
+			md_list_add(&niface->auth_md_list, md.keyid, md.key);
+			break;
+		case IMSG_RECONF_END:
+			if ((oeconf->flags & OSPFD_FLAG_STUB_ROUTER) !=
+			    (nconf->flags & OSPFD_FLAG_STUB_ROUTER))
+				stub_changed = 1;
+			else
+				stub_changed = 0;
+			merge_config(oeconf, nconf);
+			nconf = NULL;
+			if (stub_changed)
+				orig_rtr_lsa_all(NULL);
+			break;
 		case IMSG_CTL_KROUTE:
 		case IMSG_CTL_KROUTE_ADDR:
 		case IMSG_CTL_IFINFO:
@@ -327,7 +378,13 @@ ospfe_dispatch_main(int fd, short event, void *bula)
 		}
 		imsg_free(&imsg);
 	}
-	imsg_event_add(ibuf);
+	if (!shut)
+		imsg_event_add(ibuf);
+	else {
+		/* this pipe is dead, so remove the event handler */
+		event_del(&ibuf->ev);
+		event_loopexit(NULL);
+	}
 }
 
 /* ARGSUSED */
@@ -344,7 +401,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 	struct lsa_entry	*le;
 	struct imsg		 imsg;
 	struct abr_rtr		 ar;
-	int			 n, noack = 0;
+	int			 n, noack = 0, shut = 0;
 	u_int16_t		 l, age;
 
 	switch (event) {
@@ -352,7 +409,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 		if ((n = imsg_read(ibuf)) == -1)
 			fatal("imsg_read error");
 		if (n == 0)	/* connection closed */
-			fatalx("pipe closed");
+			shut = 1;
 		break;
 	case EV_WRITE:
 		if (msgbuf_write(&ibuf->w) == -1)
@@ -373,8 +430,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 		case IMSG_DD:
 			nbr = nbr_find_peerid(imsg.hdr.peerid);
 			if (nbr == NULL)
-				fatalx("ospfe_dispatch_rde: "
-				    "neighbor not found");
+				break;
 
 			/* put these on my ls_req_list for retrieval */
 			lhp = lsa_hdr_new();
@@ -384,8 +440,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 		case IMSG_DD_END:
 			nbr = nbr_find_peerid(imsg.hdr.peerid);
 			if (nbr == NULL)
-				fatalx("ospfe_dispatch_rde: "
-				    "neighbor not found");
+				break;
 
 			nbr->dd_pending--;
 			if (nbr->dd_pending == 0 && nbr->state & NBR_STA_LOAD) {
@@ -398,8 +453,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 		case IMSG_DB_SNAPSHOT:
 			nbr = nbr_find_peerid(imsg.hdr.peerid);
 			if (nbr == NULL)
-				fatalx("ospfe_dispatch_rde: "
-				    "neighbor not found");
+				break;
 
 			/* add LSA header to the neighbor db_sum_list */
 			lhp = lsa_hdr_new();
@@ -409,8 +463,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 		case IMSG_DB_END:
 			nbr = nbr_find_peerid(imsg.hdr.peerid);
 			if (nbr == NULL)
-				fatalx("ospfe_dispatch_rde: "
-				    "neighbor not found");
+				break;
 
 			/* snapshot done, start tx of dd packets */
 			nbr_fsm(nbr, NBR_EVT_SNAP_DONE);
@@ -418,8 +471,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 		case IMSG_LS_FLOOD:
 			nbr = nbr_find_peerid(imsg.hdr.peerid);
 			if (nbr == NULL)
-				fatalx("ospfe_dispatch_rde: "
-				    "neighbor not found");
+				break;
 
 			l = imsg.hdr.len - IMSG_HEADER_SIZE;
 			if (l < sizeof(lsa_hdr))
@@ -496,8 +548,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 
 			nbr = nbr_find_peerid(imsg.hdr.peerid);
 			if (nbr == NULL)
-				fatalx("ospfe_dispatch_rde: "
-				    "neighbor not found");
+				break;
 
 			if (nbr->iface->self == nbr)
 				break;
@@ -522,8 +573,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 			 */
 			nbr = nbr_find_peerid(imsg.hdr.peerid);
 			if (nbr == NULL)
-				fatalx("ospfe_dispatch_rde: "
-				    "neighbor not found");
+				break;
 
 			if (nbr->iface->self == nbr)
 				break;
@@ -548,8 +598,7 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 		case IMSG_LS_BADREQ:
 			nbr = nbr_find_peerid(imsg.hdr.peerid);
 			if (nbr == NULL)
-				fatalx("ospfe_dispatch_rde: "
-				    "neighbor not found");
+				break;
 
 			if (nbr->iface->self == nbr)
 				fatalx("ospfe_dispatch_rde: "
@@ -598,7 +647,13 @@ ospfe_dispatch_rde(int fd, short event, void *bula)
 		}
 		imsg_free(&imsg);
 	}
-	imsg_event_add(ibuf);
+	if (!shut)
+		imsg_event_add(ibuf);
+	else {
+		/* this pipe is dead, so remove the event handler */
+		event_del(&ibuf->ev);
+		event_loopexit(NULL);
+	}
 }
 
 struct iface *
@@ -673,6 +728,7 @@ orig_rtr_lsa(struct area *area)
 			rtr_link.id = iface->addr.s_addr;
 			rtr_link.data = 0xffffffff;
 			rtr_link.type = LINK_TYPE_STUB_NET;
+			rtr_link.metric = htons(iface->metric);
 			num_links++;
 			if (buf_add(buf, &rtr_link, sizeof(rtr_link)))
 				fatalx("orig_rtr_lsa: buf_add failed");
@@ -691,7 +747,12 @@ orig_rtr_lsa(struct area *area)
 				rtr_link.id = nbr->id.s_addr;
 				rtr_link.data = iface->addr.s_addr;
 				rtr_link.type = LINK_TYPE_POINTTOPOINT;
-				rtr_link.metric = htons(iface->metric);
+				/* RFC 3137: stub router support */
+				if (oeconf->flags & OSPFD_FLAG_STUB_ROUTER ||
+				    oe_nofib)
+					rtr_link.metric = 0xffff;
+				else
+					rtr_link.metric = htons(iface->metric);
 				num_links++;
 				if (buf_add(buf, &rtr_link, sizeof(rtr_link)))
 					fatalx("orig_rtr_lsa: buf_add failed");
@@ -707,8 +768,8 @@ orig_rtr_lsa(struct area *area)
 					rtr_link.id = iface->addr.s_addr;
 					rtr_link.data = iface->mask.s_addr;
 				}
-				rtr_link.metric = htons(iface->metric);
 				rtr_link.type = LINK_TYPE_STUB_NET;
+				rtr_link.metric = htons(iface->metric);
 				num_links++;
 				if (buf_add(buf, &rtr_link, sizeof(rtr_link)))
 					fatalx("orig_rtr_lsa: buf_add failed");
@@ -739,7 +800,7 @@ orig_rtr_lsa(struct area *area)
 
 			if ((iface->flags & IFF_UP) == 0 ||
 			    iface->linkstate == LINK_STATE_DOWN ||
-			    (iface->linkstate != LINK_STATE_UP &&
+			    (!LINK_STATE_IS_UP(iface->linkstate) &&
 			    iface->media_type == IFT_CARP))
 				continue;
 
@@ -761,7 +822,12 @@ orig_rtr_lsa(struct area *area)
 				rtr_link.id = nbr->id.s_addr;
 				rtr_link.data = iface->addr.s_addr;
 				rtr_link.type = LINK_TYPE_VIRTUAL;
-				rtr_link.metric = htons(iface->metric);
+				/* RFC 3137: stub router support */
+				if (oeconf->flags & OSPFD_FLAG_STUB_ROUTER ||
+				    oe_nofib)
+					rtr_link.metric = 0xffff;
+				else
+					rtr_link.metric = htons(iface->metric);
 				num_links++;
 				virtual = 1;
 				if (buf_add(buf, &rtr_link, sizeof(rtr_link)))
@@ -777,6 +843,7 @@ orig_rtr_lsa(struct area *area)
 			rtr_link.id = iface->addr.s_addr;
 			rtr_link.data = 0xffffffff;
 			rtr_link.type = LINK_TYPE_STUB_NET;
+			rtr_link.metric = htons(iface->metric);
 			num_links++;
 			if (buf_add(buf, &rtr_link, sizeof(rtr_link)))
 				fatalx("orig_rtr_lsa: buf_add failed");
@@ -786,12 +853,18 @@ orig_rtr_lsa(struct area *area)
 				    nbr->state & NBR_STA_FULL) {
 					bzero(&rtr_link, sizeof(rtr_link));
 					log_debug("orig_rtr_lsa: "
-					    "point-to-point, interface %s",
+					    "point-to-multipoint, interface %s",
 					    iface->name);
 					rtr_link.id = nbr->addr.s_addr;
 					rtr_link.data = iface->addr.s_addr;
 					rtr_link.type = LINK_TYPE_POINTTOPOINT;
-					rtr_link.metric = htons(iface->metric);
+					/* RFC 3137: stub router support */
+					if (oe_nofib || oeconf->flags &
+					    OSPFD_FLAG_STUB_ROUTER)
+						rtr_link.metric = 0xffff;
+					else
+						rtr_link.metric =
+						    htons(iface->metric);
 					num_links++;
 					if (buf_add(buf, &rtr_link,
 					    sizeof(rtr_link)))
@@ -805,7 +878,12 @@ orig_rtr_lsa(struct area *area)
 		}
 
 		rtr_link.num_tos = 0;
-		rtr_link.metric = htons(iface->metric);
+		/* RFC 3137: stub router support */
+		if ((oeconf->flags & OSPFD_FLAG_STUB_ROUTER || oe_nofib) &&
+		    rtr_link.type != LINK_TYPE_STUB_NET)
+			rtr_link.metric = 0xffff;
+		else
+			rtr_link.metric = htons(iface->metric);
 		num_links++;
 		if (buf_add(buf, &rtr_link, sizeof(rtr_link)))
 			fatalx("orig_rtr_lsa: buf_add failed");
@@ -828,11 +906,11 @@ orig_rtr_lsa(struct area *area)
 
 	if (oeconf->border)
 		lsa_rtr.flags |= OSPF_RTR_B;
+	/* TODO set V flag if a active virtual link ends here and the
+	 * area is the tranist area for this link. */
 	if (virtual)
 		lsa_rtr.flags |= OSPF_RTR_V;
 
-	/* TODO set V flag if a active virtual link ends here and the
-	 * area is the tranist area for this link. */
 	lsa_rtr.dummy = 0;
 	lsa_rtr.nlinks = htons(num_links);
 	memcpy(buf_seek(buf, sizeof(lsa_hdr), sizeof(lsa_rtr)),
@@ -929,6 +1007,19 @@ ospfe_router_id(void)
 	return (oeconf->rtr_id.s_addr);
 }
 
+void
+ospfe_fip_update(int type)
+{
+	int	old = oe_nofib;
+
+	if (type == IMSG_CTL_FIB_COUPLE)
+		oe_nofib = 0;
+	if (type == IMSG_CTL_FIB_DECOUPLE)
+		oe_nofib = 1;
+	if (old != oe_nofib)
+		orig_rtr_lsa_all(NULL);
+}
+
 void
 ospfe_iface_ctl(struct ctl_conn *c, unsigned int idx)
 {
@@ -966,3 +1057,42 @@ ospfe_nbr_ctl(struct ctl_conn *c)
 
 	imsg_compose(&c->ibuf, IMSG_CTL_END, 0, 0, NULL, 0);
 }
+
+void
+ospfe_demote_area(struct area *area, int active)
+{
+	struct demote_msg	dmsg;
+
+	if (ospfd_process != PROC_OSPF_ENGINE ||
+	    area->demote_group[0] == '\0')
+		return;
+
+	bzero(&dmsg, sizeof(dmsg));
+	strlcpy(dmsg.demote_group, area->demote_group,  
+	sizeof(dmsg.demote_group));
+	dmsg.level = area->demote_level;
+	if (active)
+		dmsg.level = -dmsg.level;
+
+	ospfe_imsg_compose_parent(IMSG_DEMOTE, 0, &dmsg, sizeof(dmsg));
+}
+
+void
+ospfe_demote_iface(struct iface *iface, int active)
+{
+	struct demote_msg	dmsg;
+
+	if (ospfd_process != PROC_OSPF_ENGINE ||
+	    iface->demote_group[0] == '\0')
+		return;
+
+	bzero(&dmsg, sizeof(dmsg));
+	strlcpy(dmsg.demote_group, iface->demote_group,  
+	sizeof(dmsg.demote_group));
+	if (active)
+		dmsg.level = -1;
+	else
+		dmsg.level = 1;
+
+	ospfe_imsg_compose_parent(IMSG_DEMOTE, 0, &dmsg, sizeof(dmsg));
+}
