@@ -1,4 +1,4 @@
-/*	$OpenBSD: ospfe.h,v 1.31 2006/06/02 18:49:55 norby Exp $ */
+/*	$OpenBSD: ospfe.h,v 1.36 2007/05/29 22:08:25 claudio Exp $ */
 
 /*
  * Copyright (c) 2004, 2005 Esben Norby <norby@openbsd.org>
@@ -100,6 +100,7 @@ int		 auth_gen(struct buf *, struct iface *);
 void		 md_list_add(struct auth_md_head *, u_int8_t, char *);
 void		 md_list_copy(struct auth_md_head *, struct auth_md_head *);
 void		 md_list_clr(struct auth_md_head *);
+int		 md_list_send(struct auth_md_head *, struct imsgbuf *);
 
 /* database.c */
 int	 send_db_description(struct nbr *);
@@ -124,15 +125,18 @@ int		 ospfe_imsg_compose_parent(int, pid_t, void *, u_int16_t);
 int		 ospfe_imsg_compose_rde(int, u_int32_t, pid_t, void *,
 		     u_int16_t);
 u_int32_t	 ospfe_router_id(void);
+void		 ospfe_fip_update(int);
 void		 ospfe_iface_ctl(struct ctl_conn *, unsigned int);
 void		 ospfe_nbr_ctl(struct ctl_conn *);
 void		 orig_rtr_lsa(struct area *);
 void		 orig_net_lsa(struct iface *);
+void		 ospfe_demote_area(struct area *, int);
+void		 ospfe_demote_iface(struct iface *, int);
 
 /* interface.c */
 int		 if_fsm(struct iface *, enum iface_event);
 
-struct iface	*if_new(struct kif *);
+struct iface	*if_new(struct kif *, struct kif_addr *);
 void		 if_del(struct iface *);
 void		 if_init(struct ospfd_conf *, struct iface *);
 
@@ -145,10 +149,10 @@ struct ctl_iface	*if_to_ctl(struct iface *);
 int	 if_join_group(struct iface *, struct in_addr *);
 int	 if_leave_group(struct iface *, struct in_addr *);
 int	 if_set_mcast(struct iface *);
-int	 if_set_mcast_ttl(int, u_int8_t);
-int	 if_set_tos(int, int);
+int	 if_set_recvif(int, int);
 void	 if_set_recvbuf(int);
 int	 if_set_mcast_loop(int);
+int	 if_set_ip_hdrincl(int);
 
 /* lsack.c */
 int	 delay_lsa_ack(struct iface *, struct lsa_hdr *);
