@@ -1,4 +1,4 @@
-/*	$OpenBSD: rde_lsdb.c,v 1.33 2006/08/30 05:25:33 norby Exp $ */
+/*	$OpenBSD: rde_lsdb.c,v 1.38 2007/04/10 13:26:39 claudio Exp $ */
 
 /*
  * Copyright (c) 2004, 2005 Claudio Jeker <claudio@openbsd.org>
@@ -69,7 +69,9 @@ vertex_get(struct lsa *lsa, struct rde_nbr *nbr)
 
 	if ((v = calloc(1, sizeof(struct vertex))) == NULL)
 		fatal(NULL);
-	v->nbr = nbr;
+	TAILQ_INIT(&v->nexthop);
+	v->area = nbr->area;
+	v->peerid = nbr->peerid;
 	v->lsa = lsa;
 	clock_gettime(CLOCK_MONOTONIC, &tp);
 	v->changed = v->stamp = tp.tv_sec;
@@ -77,8 +79,10 @@ vertex_get(struct lsa *lsa, struct rde_nbr *nbr)
 	v->ls_id = ntohl(lsa->hdr.ls_id);
 	v->adv_rtr = ntohl(lsa->hdr.adv_rtr);
 	v->type = lsa->hdr.type;
+
 	if (!nbr->self)
 		v->flooded = 1; /* XXX fix me */
+	v->self = nbr->self;
 
 	evtimer_set(&v->ev, lsa_timeout, v);
 
@@ -91,7 +95,7 @@ vertex_free(struct vertex *v)
 	if (v->type == LSA_TYPE_EXTERNAL)
 		RB_REMOVE(lsa_tree, &asext_tree, v);
 	else
-		RB_REMOVE(lsa_tree, &v->nbr->area->lsa_tree, v);
+		RB_REMOVE(lsa_tree, &v->area->lsa_tree, v);
 
 	(void)evtimer_del(&v->ev);
 	free(v->lsa);
@@ -353,9 +357,6 @@ lsa_add(struct rde_nbr *nbr, struct lsa *lsa)
 			RB_INSERT(lsa_tree, tree, new);
 			new->deleted = 1;
 
-			log_debug("lsa_add: removing %p and putting %p on hold "
-			    "queue (timeout %d)", old, new, res.tv_sec);
-
 			if (evtimer_add(&new->ev, &res) != 0)
 				fatal("lsa_add");
 			return (1);
@@ -396,14 +397,12 @@ lsa_del(struct rde_nbr *nbr, struct lsa_hdr *lsa)
 	if (v == NULL)
 		return;
 
-	log_debug("lsa_del: putting %p on hold queue", v);
-
 	v->deleted = 1;
 	/* hold time to make sure that a new lsa is not added premature */
 	timerclear(&tv);
 	tv.tv_sec = MIN_LS_INTERVAL;
 	if (evtimer_add(&v->ev, &tv) == -1)
-		fatal("lsa_merge");
+		fatal("lsa_del");
 }
 
 void
@@ -563,7 +562,6 @@ lsa_dump(struct lsa_tree *tree, int imsg_type, pid_t pid)
 				break;
 			continue;
 		default:
-			log_debug("%d", imsg_type);
 			log_warnx("lsa_dump: unknown imsg type");
 			return;
 		}
@@ -583,39 +581,35 @@ lsa_timeout(int fd, short event, void *bula)
 
 	if (v->deleted) {
 		if (ntohs(v->lsa->hdr.age) >= MAX_AGE) {
-			log_debug("lsa_timeout: finally free %p", v);
 			vertex_free(v);
 		} else {
 			v->deleted = 0;
 
 			/* schedule recalculation of the RIB */
 			if (v->lsa->hdr.type != LSA_TYPE_EXTERNAL)
-				v->nbr->area->dirty = 1;
+				v->area->dirty = 1;
 			start_spf_timer();
 
-			rde_imsg_compose_ospfe(IMSG_LS_FLOOD, v->nbr->peerid, 0,
+			rde_imsg_compose_ospfe(IMSG_LS_FLOOD, v->peerid, 0,
 			    v->lsa, ntohs(v->lsa->hdr.len));
 
 			/* timeout handling either MAX_AGE or LS_REFRESH_TIME */
 			timerclear(&tv);
-			if (v->nbr->self)
+			if (v->self)
 				tv.tv_sec = LS_REFRESH_TIME;
 			else
 				tv.tv_sec = MAX_AGE - ntohs(v->lsa->hdr.age);
 
-			log_debug("lsa_timeout: flodding %p new timeout %d",
-			    v, tv.tv_sec);
-
 			if (evtimer_add(&v->ev, &tv) != 0)
 				fatal("lsa_add");
 		}
 		return;
 	}
 
-	if (v->nbr->self && ntohs(v->lsa->hdr.age) < MAX_AGE)
+	if (v->self && ntohs(v->lsa->hdr.age) < MAX_AGE)
 		lsa_refresh(v);
 
-	rde_imsg_compose_ospfe(IMSG_LS_FLOOD, v->nbr->peerid, 0,
+	rde_imsg_compose_ospfe(IMSG_LS_FLOOD, v->peerid, 0,
 	    v->lsa, ntohs(v->lsa->hdr.len));
 }
 
@@ -707,7 +701,8 @@ lsa_remove_invalid_sums(struct area *area)
 		nv = RB_NEXT(lsa_tree, tree, v);
 		if ((v->lsa->hdr.type == LSA_TYPE_SUM_NETWORK ||
 		    v->lsa->hdr.type == LSA_TYPE_SUM_ROUTER) &&
-		    v->nbr->self && v->cost == LS_INFINITY) {
+		    v->self && v->cost == LS_INFINITY &&
+		    v->deleted == 0) {
 			/*
 			 * age the lsa and call lsa_timeout() which will
 			 * actually remove it from the database.
